/************************************************************************************
RMP RT-Component
Copyright (c) 2010, Segway Japan, Ltd.
All rights reserved.

Contact us if you use this software for sell.
If you use this software not for sell, you can use this software under BSD lisence.
See the files LICENSE.TXT and LICENSE-BSD.TXT for more details.                     
************************************************************************************/
#include <math.h>
#include "Rmp.h"
#include "RmpTEST.h"

// Module specification
// <rtc-template block="module_spec">
static const char* rmp_spec[] =
  {
    "implementation_id", "Rmp",
    "type_name",         "Rmp",
    "description",       "Rmp Component",
    "version",           "0.1.0",
    "vendor",            "SegwayJapan",
    "category",          "Generic",
    "activity_type",     "SPORADIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "10",
    "language",          "C++",
    "lang_type",         "compile",
    //"exec_cxt.periodic.rate", "1.0",
    // Configuration variables
    "conf.default.Kspeed", "1.0",
    "conf.default.KspeedA", "-0.01630197",
    "conf.default.KspeedB", "1.412417763",
    "conf.default.Kturn", "1.0",
    "conf.default.wheelwidth", "0.54",
    "conf.default.wheelradiusL", "0.235",
    "conf.default.wheelradiusR", "0.235",
    "conf.default.wheelcount1rot", "49646.0775",
    "conf.default.str_port", "FTDI",
    "conf.default.Kodoxy", "1.0",
    "conf.default.Kodotheta", "1.0",
    "conf.default.simulator", "0",
    "conf.default.idIIS", "200",
    "conf.default.startxIIS", "0.0",
    "conf.default.startyIIS", "0.0",
    "conf.default.startthetaIIS", "0.0",
    "conf.default.Mode", "0",
    "conf.default.MaxV", "2.0",
    "conf.default.MaxW", "2.0",
    ""
  };
// </rtc-template>

Rmp::Rmp(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_InputIn("Input", m_Input),
    m_VelocityIn("Velocity", m_Velocity),
    m_VelocityIISIn("VelocityIIS", m_VelocityIIS),
    m_Velocity2DIISIn("Velocity2DIIS", m_Velocity2DIIS),
    m_EncoderLOut("EncoderL", m_EncoderL),
    m_EncoderROut("EncoderR", m_EncoderR),
    m_BatteryOut("Battery", m_Battery),
    m_UiBatteryOut("UiBattery", m_UiBattery),
    m_PitchAngleOut("PitchAngle", m_PitchAngle),
    m_PitchRateOut("PitchRate", m_PitchRate),
    m_RollAngleOut("RollAngle", m_RollAngle),
    m_RollRateOut("RollRate", m_RollRate),
    m_YawRateOut("YawRate", m_YawRate),
    m_OdometryOut("Odometry", m_Odometry),
    m_OdometryBOut("OdometryB", m_OdometryB),
    m_RobotVelIISOut("RobotVelIIS", m_RobotVelIIS),
    m_RobotPosIISOut("RobotPosIIS", m_RobotPosIIS),
    m_RobotVel2DIISOut("RobotVel2DIIS", m_RobotVel2DIIS),
    m_RobotPose2DIISOut("RobotPose2DIIS", m_RobotPose2DIIS),

    // </rtc-template>
	dummy(0)
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  registerInPort("Input", m_InputIn);
  registerInPort("Velocity", m_VelocityIn);
  registerInPort("VelocityIIS", m_VelocityIISIn);
  registerInPort("Velocity2DIIS", m_Velocity2DIISIn);
  
  // Set OutPort buffer
  registerOutPort("EncoderL", m_EncoderLOut);
  registerOutPort("EncoderR", m_EncoderROut);
  registerOutPort("Battery", m_BatteryOut);
  registerOutPort("UiBattery", m_UiBatteryOut);
  registerOutPort("PitchAngle", m_PitchAngleOut);
  registerOutPort("PitchRate", m_PitchRateOut);
  registerOutPort("RollAngle", m_RollAngleOut);
  registerOutPort("RollRate", m_RollRateOut);
  registerOutPort("YawRate", m_YawRateOut);
  registerOutPort("Odometry", m_OdometryOut);
  registerOutPort("OdometryB", m_OdometryBOut);
  registerOutPort("RobotVelIIS", m_RobotVelIISOut);
  registerOutPort("RobotPosIIS", m_RobotPosIISOut);
  registerOutPort("RobotVel2DIIS", m_RobotVel2DIISOut);
  registerOutPort("RobotPose2DIIS", m_RobotPose2DIISOut);
  
  // Set service provider to Ports
  
  // Set service consumers to Ports
  
  // Set CORBA Service Ports
  
  // </rtc-template>

}

Rmp::~Rmp()
{
}


RTC::ReturnCode_t Rmp::onInitialize()
{
  // <rtc-template block="bind_config">
  // Bind variables and configuration variable
  bindParameter("Kspeed", m_Kspeed, "1.0");
  bindParameter("KspeedA", m_KspeedA, "-0.01630197");
  bindParameter("KspeedB", m_KspeedB, "1.412417763");
  bindParameter("Kturn", m_Kturn, "1.0");
  bindParameter("wheelwidth", m_wheelwidth, "0.54");
  bindParameter("wheelradiusL", m_wheelradiusL, "0.235");
  bindParameter("wheelradiusR", m_wheelradiusR, "0.235");
  bindParameter("wheelcount1rot", m_wheelcount1rot, "49646.0775");
  bindParameter("str_port", m_str_port, "FTDI");
  bindParameter("Kodoxy", m_Kodoxy, "1.0");
  bindParameter("Kodotheta", m_Kodotheta, "1.0");
  bindParameter("simulator", m_simulator, "0");
  bindParameter("idIIS", m_idIIS, "200");
  bindParameter("startxIIS", m_startxIIS, "0.0");
  bindParameter("startyIIS", m_startyIIS, "0.0");
  bindParameter("startthetaIIS", m_startthetaIIS, "0.0");
  bindParameter("Mode", m_Mode, "0");
  bindParameter("MaxV", m_MaxV, "2.0");
  bindParameter("MaxW", m_MaxW, "2.0");
  bindParameter("MaxW", m_MaxW, "2.0");
  
  // </rtc-template>



  return RTC::RTC_OK;
}



/*
RTC::ReturnCode_t Rmp::onFinalize()
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t Rmp::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t Rmp::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


RTC::ReturnCode_t Rmp::onActivated(RTC::UniqueId ec_id)
{
#if !RMP_TEST04_NO_RMP

#if !RMP_TEST03_NO_RMPIF
  rmp_open((char*)m_str_port.c_str(), Status);
#endif

  Status.wheelwidth   = m_wheelwidth;
  Status.wheelradiusL = m_wheelradiusL;
  Status.wheelradiusR = m_wheelradiusR;
  Status.count1rot    = m_wheelcount1rot;

  Status.odox     = -m_startyIIS;
  Status.odoy     =  m_startxIIS;
  Status.odotheta =  m_startthetaIIS;

  Stop = 0;
  Speed = 0;
  Turn = 0;


#define USE_RMP_MAXFACTOR (0)
#if USE_RMP_MAXFACTOR
    //------------------------------------------------------------
    //Speed max=1176 Turn max=1024
    //max speedmps = 1176 / 332 = 3.542168675              RMP Interface Guide p7 and p13
    //max turnrps  = 1024 / 446.90717 = 2.291303583        RMP Interface Guide p7 and p13
    //-----------------------------
    //what is max v(mps)?
    //-----------------------------
    //when max velocity factor = 0.75
    //1176 = a * 332.0  
    //a = 3.542168675
    //3.542168675 = v * (-0.0163019 * v + 1.41241776)
    //0 = -0.0163019 * v^2 + 1.41241776 * v - 3.542168675
    //v = 2.585001     (max mps max velocity factor = 0.75)
    //v*(4/3)= 3.446666(max mps max velocity factor = 1.0)
    //------------------------------------------------------------
    double Maxspeedmps = 3.446666;
    double Maxturnrps  = 1024.0 / 446.90717; //2.2913
    double MaxVelocityScaleFactor = m_MaxV / Maxspeedmps; //0.0 - 1.0
    double MaxTurnRateScaleFactor = m_MaxW / Maxturnrps;  //0.0 - 1.0
    MaxVelocityScaleFactor16 = (unsigned short)(16.0 * MaxVelocityScaleFactor); //0-16
    MaxTurnRateScaleFactor16 = (unsigned short)(16.0 * MaxTurnRateScaleFactor); //0-16

    if (m_MaxW > 0.0) {
      if (MaxTurnRateScaleFactor16 > 16) {
	MaxTurnRateScaleFactor16 = 16;
      }
      Kturnrps  = 16.0 / (double)MaxTurnRateScaleFactor16;

      printf("#MaxTurnRateScaleFactor = %f\n", (double)MaxTurnRateScaleFactor16/16.0); 
      rmp_setconfiguration(COMMAND_SET_MAXIMUM_TURN_RATE_SCALE_FACTOR, MaxTurnRateScaleFactor16);
    }

    if (m_MaxV > 0.0) {
      if (MaxVelocityScaleFactor16 > 16) {
	MaxVelocityScaleFactor16 = 16;
      }
      Kspeedmps = 16.0 / (double)MaxVelocityScaleFactor16;
      
      printf("#MaxVelocityScaleFactor = %f\n", (double)MaxVelocityScaleFactor16/16.0);
      rmp_setconfiguration(COMMAND_SET_MAX_VELOCITY_SCALE_FACTOR, MaxVelocityScaleFactor16);
    }
#endif


#endif



  return RTC::RTC_OK;
}



RTC::ReturnCode_t Rmp::onDeactivated(RTC::UniqueId ec_id)
{
#if !RMP_TEST04_NO_RMP
#if !RMP_TEST03_NO_RMPIF
  rmp_close();
#endif
#endif
  return RTC::RTC_OK;
}


/*!
  @brief send speed,turn rate to RMP
  @brief recv encoder, battery, pitchangle status from RMP
  @return
*/

#ifndef WIN32
#include <time.h>
#include <sys/time.h>

unsigned long TimeInterval()
{
  static bool first = true;
  static unsigned long long tvprev;

  timeval tv;

  if (first) {
    gettimeofday(&tv, NULL);
    tvprev = tv.tv_sec*1000000 + tv.tv_usec;
    first = false;
  }
    gettimeofday(&tv, NULL);

    unsigned long long tvcurr, tvdiff;
    tvcurr = tv.tv_sec*1000000 + tv.tv_usec;
    tvdiff = tvcurr - tvprev;
    tvprev = tvcurr;

    return (unsigned long)tvdiff/1000;
}
void TimeAverage()
{
    static float time[128];
    static int i = 0;

    time[i] = (float)TimeInterval();

    i++;
    if (i == 128) {
      float sum = 0;
      for(int j = 0; j < 128; j++) {
	sum += time[j];
      }
      sum /= 128.0f;
      i = 0;
      printf("#use msec %f\n", sum);
    }
}

#endif


RTC::ReturnCode_t Rmp::onExecute(RTC::UniqueId ec_id)
{

#if !RMP_TEST04_NO_RMP

  if (m_InputIn.isNew()) {
    while (!m_InputIn.isEmpty()) {
      m_InputIn.read();
    }
    Stop = m_Input.data;
  }

  if (m_VelocityIn.isNew() || 
      m_VelocityIISIn.isNew() || 
      m_Velocity2DIISIn.isNew()) {

    double speedmps = 0.0;
    double turnrps = 0.0;


    if (m_VelocityIn.isNew()) {
      while (!m_VelocityIn.isEmpty()) {
	m_VelocityIn.read();
      }

      speedmps = m_Velocity.v;
      turnrps = m_Velocity.w;
    }
    if (m_VelocityIISIn.isNew()) {
      while (!m_VelocityIISIn.isEmpty()) {
	m_VelocityIISIn.read();
      }

      speedmps = m_VelocityIIS.vx;
      turnrps = m_VelocityIIS.w;
    }
    if (m_Velocity2DIISIn.isNew()) {
      while (!m_Velocity2DIISIn.isEmpty()) {
	m_Velocity2DIISIn.read();
      }

      speedmps = m_Velocity2DIIS.data.vx;
      turnrps = m_Velocity2DIIS.data.va;
    }

    //printf("iv %f %f    ", speedmps, turnrps);
   
    //VELOCITY
    //V = V * (A * |V| + B) * Kspeed
    speedmps = speedmps * (m_KspeedA * fabs((float)speedmps) + m_KspeedB);
    speedmps *= m_Kspeed;

    if (m_MaxV > 0.0 && fabs(speedmps) > m_MaxV) {
      speedmps = (speedmps > 0.0)? m_MaxV : -m_MaxV;
      printf("#V was limited to MaxV\n");
    }
    Speed = (int) (speedmps * 332.0);   //RMP Interface Guide p13

    //TURN
    turnrps *= m_Kturn;

    if (m_MaxW > 0.0 && fabs(turnrps) > m_MaxW) {
      turnrps = (turnrps > 0.0)? m_MaxW : -m_MaxW;
      printf("#W was limited to MaxW\n");
    }
    Turn = (int) (turnrps * 180.0 / 3.141592654  * 7.8); //RMP Interface Guide p13
  }


  if (Stop) {
    Speed = 0;
    Turn = 0;
  }


#if USE_RMP_MAXFACTOR
  unsigned short command = 0;
  unsigned short param   = 0;
  
  static unsigned int cnt = 0;
  cnt++;
  if (cnt & 1) {
    command = COMMAND_SET_MAX_VELOCITY_SCALE_FACTOR;
    param = MaxVelocityScaleFactor16;
  }
  else {
    command = COMMAND_SET_MAXIMUM_TURN_RATE_SCALE_FACTOR;
    param = MaxTurnRateScaleFactor16;
  }
  rmp_setspeed((short)Speed, (short)Turn, command, param);
#else

#if !RMP_TEST03_NO_RMPIF
  rmp_setspeed((short)Speed, (short)Turn);
#endif
#endif




#if !RMP_TEST03_NO_RMPIF
  rmp_getstatus(Status, m_Kodoxy, m_Kodotheta);
#endif





  //output velocity
  if (m_simulator == 0 ||
      fabs((float)Status.addx) > 0.0 ||
      fabs((float)Status.addy) > 0.0 || 
      fabs((float)Status.addtheta) > 0.0) {

   //culc speed
    coil::TimeValue tdiff = Status.odotime - Status.odotime_prev;
    double sec    = (double)((double)tdiff.sec() + (double)tdiff.usec() / 1000000.0);
    double dx     = Status.addx / sec;     //(m/sec)
    double dy     = Status.addy / sec;     //(m/sec)
    double dtheta = Status.addtheta / sec; //(rad/sec)

    //SJ to IIS coord
    double dxIIS = dy;
    double dyIIS = -dx;
    double dthetaIIS = dtheta;

    m_RobotVelIIS.id.length(1);
    m_RobotVelIIS.id[0]   = m_idIIS;
    m_RobotVelIIS.tm.sec  = (::CORBA::ULong) Status.odotime.sec();
    m_RobotVelIIS.tm.nsec = (::CORBA::ULong) (Status.odotime.usec() * 1000);
    m_RobotVelIIS.vx      = dxIIS;
    m_RobotVelIIS.vy      = dyIIS;
    m_RobotVelIIS.w       = dthetaIIS;
    
    m_RobotVel2DIIS.id.length(1);
    m_RobotVel2DIIS.id[0]   = m_idIIS;
    m_RobotVel2DIIS.tm.sec  = (::CORBA::ULong) Status.odotime.sec();
    m_RobotVel2DIIS.tm.nsec = (::CORBA::ULong) (Status.odotime.usec() * 1000);
    m_RobotVel2DIIS.data.vx = dxIIS;
    m_RobotVel2DIIS.data.vy = dyIIS;
    m_RobotVel2DIIS.data.va = dthetaIIS;

    //printf("ov %f %f %f\n", m_RobotVel2DIIS.data.vx, m_RobotVel2DIIS.data.vy, m_RobotVel2DIIS.data.va);
    
  }

  {
    //SJ to IIS coord
    m_RobotPosIIS.id.length(1);
    m_RobotPosIIS.id[0]   = m_idIIS;
    m_RobotPosIIS.tm.sec  = (::CORBA::ULong) Status.odotime.sec();
    m_RobotPosIIS.tm.nsec = (::CORBA::ULong) (Status.odotime.usec() * 1000);
    m_RobotPosIIS.x       = Status.odoy;
    m_RobotPosIIS.y       = -Status.odox;
    m_RobotPosIIS.theta   = Status.odotheta;

    m_RobotPose2DIIS.id.length(1);
    m_RobotPose2DIIS.id[0]   = m_idIIS;
    m_RobotPose2DIIS.tm.sec  = (::CORBA::ULong) Status.odotime.sec();
    m_RobotPose2DIIS.tm.nsec = (::CORBA::ULong) (Status.odotime.usec() * 1000);
    m_RobotPose2DIIS.data.position.x       = Status.odoy;
    m_RobotPose2DIIS.data.position.y       = -Status.odox;
    m_RobotPose2DIIS.data.heading          = Status.odotheta;

    //outodo
    //printf("pos %f %f %f\n", m_RobotPosIIS.x, m_RobotPosIIS.y, m_RobotPosIIS.theta);

    {
      //test
      static double prevx,prevy;
      double distance = sqrt(
			     (m_RobotPosIIS.x - prevx)*(m_RobotPosIIS.x - prevx) + 
			     (m_RobotPosIIS.y - prevy)*(m_RobotPosIIS.y - prevy));
      if (distance > 10.0) {
	printf("#OVER ODO (%f %f)-(%f %f) %f \n", prevx, prevy, m_RobotPosIIS.x, m_RobotPosIIS.y, distance);
      }
      prevx = m_RobotPosIIS.x;
      prevy = m_RobotPosIIS.y;
    }
  }

  RTC::Time tm401;
  tm401.sec  = (::CORBA::ULong) Status.tv401.sec();
  tm401.nsec = (::CORBA::ULong) (Status.tv401.usec() * 1000);

  RTC::Time tm402;
  tm402.sec  = (::CORBA::ULong) Status.tv402.sec();
  tm402.nsec = (::CORBA::ULong) (Status.tv402.usec() * 1000);

  RTC::Time tm403;
  tm403.sec  = (::CORBA::ULong) Status.tv403.sec();
  tm403.nsec = (::CORBA::ULong) (Status.tv403.usec() * 1000);

  RTC::Time tm406;
  tm406.sec  = (::CORBA::ULong) Status.tv406.sec();
  tm406.nsec = (::CORBA::ULong) (Status.tv406.usec() * 1000);

  RTC::Time tmodo;
  tmodo.sec  = (::CORBA::ULong) Status.odotime.sec();
  tmodo.nsec = (::CORBA::ULong) (Status.odotime.usec() * 1000);

  m_EncoderL.tm     = tm403;
  m_EncoderL.data   = Status.int_left_wheel;
  m_EncoderR.tm     = tm403;
  m_EncoderR.data   = Status.int_right_wheel;

  m_Battery.tm      = tm406;
  m_Battery.data    = Status.battery_voltage;
  m_UiBattery.tm    = tm406;
  m_UiBattery.data  = Status.ui_battery_voltage;
  
  m_PitchAngle.tm   = tm401;
  m_PitchAngle.data = Status.pitchangle;
  m_PitchRate.tm    = tm401;
  m_PitchRate.data  = Status.pitchrate;
  m_RollAngle.tm    = tm401;
  m_RollAngle.data  = Status.rollangle;
  m_RollRate.tm     = tm401;
  m_RollRate.data   = Status.rollrate;

  m_YawRate.tm      = tm402;
  m_YawRate.data    = Status.yawrate;

  m_Odometry.tm     = tmodo;
  m_Odometry.x      = Status.odox;
  m_Odometry.y      = Status.odoy;
  m_Odometry.theta  = Status.odotheta;

  m_OdometryB.tm    = tmodo;
  m_OdometryB.x     = m_RobotPosIIS.x;
  m_OdometryB.y     = m_RobotPosIIS.y;
  m_OdometryB.theta = m_RobotPosIIS.theta;


  //warning
  if (Status.battery_voltage < 60.0) {
    static int warning = 0;
    warning++;
    if (warning % 30) {
      printf("#Warning Battery Voltage %f\n", Status.battery_voltage);
    }
  }


  //  printf("stat Speed/Turn %d %d EncLR %d %d Batt/UIBatt %d %d\n",
  //      (int)m_Speed.data, (int)m_Turn.data, (int)m_EncoderL.data, (int)m_EncoderR.data, (int) m_Battery.data, (int) m_UiBattery.data);

  m_EncoderLOut.write();
  m_EncoderROut.write();
  m_BatteryOut.write();
  m_UiBatteryOut.write();
  m_PitchAngleOut.write();
  m_PitchRateOut.write();
  m_RollAngleOut.write();
  m_RollRateOut.write();
  m_YawRateOut.write();
  m_OdometryOut.write();
  m_OdometryBOut.write();


  //IIS 
  m_RobotVelIISOut.write();
  m_RobotVel2DIISOut.write();

  switch (m_Mode) {
  case 1:
    if (Stop) {//Stop used as pushed button
      printf("#write RobotPOSIIS %f %f %f\n", m_RobotPosIIS.x, m_RobotPosIIS.y, m_RobotPosIIS.theta);
      m_RobotPosIISOut.write();
      m_RobotPose2DIISOut.write();
    }
    break;
  case 0:
    m_RobotPosIISOut.write();
    m_RobotPose2DIISOut.write();
    break;
  }


#endif

  //  usleep(100);

  return RTC::RTC_OK;
}


/*
RTC::ReturnCode_t Rmp::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t Rmp::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t Rmp::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t Rmp::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t Rmp::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/



extern "C"
{

  void RmpInit(RTC::Manager* manager)
  {
    RTC::Properties profile(rmp_spec);
    manager->registerFactory(profile,
                             RTC::Create<Rmp>,
                             RTC::Delete<Rmp>);
  }

};


