/************************************************************************************
RMP RT-Component
Copyright (c) 2010, Segway Japan, Ltd.
All rights reserved.

Contact us if you use this software for sell.
If you use this software not for sell, you can use this software under BSD lisence.
See the files LICENSE.TXT and LICENSE-BSD.TXT for more details.                     
************************************************************************************/
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "Rmp.h"
#include "rmp_if.h"
#include "usb_io.h"

#include "RmpTEST.h"


#define PRINTOUT (0)

char device[128] = "FTDI";



unsigned int readloopcnt = 3;

void rmp_setport(char* port)
{
  if (port) {
    
    readloopcnt = 1;
    if (strcmp(port, "FTDI") == 0) {
      readloopcnt = 3;
    }

    if (strlen(port) > 0) {
      strcpy(device, port);
    }
  }
}

void rmp_open(char* port, rmp_status& status)
{
  status.odox = 0.0;
  status.odoy = 0.0;
  status.odotheta = 0.0;

  status.odogot = false;
  status.odofirst = true;

  rmp_setport(port);

#if !RMP_TEST02_NO_FTDI
  usbs_init();
#endif
  
  status.tv401 = 
  status.tv402 = 
  status.tv403 = 
  status.tv406 = 
  status.odotime = 
  status.odotime_prev = coil::gettimeofday();

}

void rmp_close()
{
#if !RMP_TEST02_NO_FTDI
  usbs_Close();
#endif
}


//===========================================
//STATUS
//==========================================

void rmp_getstatus_sub(unsigned char* statusbuf, rmp_status& status)
{
  unsigned char* bytes = statusbuf;
  int id = ((bytes[4] << 3) | ((bytes[5] >> 5) & 7)) & 0x0fff;

#if 0
  printf("id %x:", id);
  for (int i = 0; i < 18; i++) {
    printf("%02x ", bytes[i]);
  }
  printf("\n");
#endif

  short data[4];
  data[0] = (bytes[9]  << 8) | bytes[10];
  data[1] = (bytes[11] << 8) | bytes[12];
  data[2] = (bytes[13] << 8) | bytes[14];
  data[3] = (bytes[15] << 8) | bytes[16];




  switch (id) {
  case 0x401:
    status.tv401      = coil::gettimeofday();
    status.pitchangle = (float)data[0] / 7.8f;
    status.pitchrate  = (float)data[1] / 7.8f;
    status.rollangle  = (float)data[2] / 7.8f;
    status.rollrate   = (float)data[3] / 7.8f;
    break;
  case 0x402:
    status.tv402      = coil::gettimeofday();
    status.lwvel      = (float)data[0] / 332.0f;
    status.rwvel      = (float)data[1] / 332.0f;
    status.yawrate    = (float)data[2] / 7.8f;
    break;
  case 0x403:
    status.tv403           = coil::gettimeofday();
    status.int_left_wheel  = (data[1] << 16) | (data[0] & 0x0ffff);
    status.int_right_wheel = (data[3] << 16) | (data[2] & 0x0ffff);
    status.odogot = true;
#if PRINTOUT
    printf("%d %d\n", status.int_left_wheel, status.int_right_wheel); 
#endif
    break;

  case 0x404:
    break;

  case 0x405:
    break;

  case 0x406:
    status.tv406              = coil::gettimeofday();
    status.ui_battery_voltage = (1.5 + ((float)data[2] * 0.0125));
    status.battery_voltage    = ((float)data[3]) / (float)4.0; 
#if PRINTOUT
    printf("%f %f\n", status.ui_battery_voltage, status.battery_voltage);
#endif
    break;

  case 0x407:
#if 0
    status.velocity = data[0];
    status.turn = data[1];
#if PRINTOUT
    printf("%d %d\n", status.velocity, status.turn);
#endif
#endif
    break;
  }



}



bool rmp_odometry(rmp_status& status, double kxy, double ktheta)
{
#define RMP200_WHEELR  (0.47*0.5)
#define RMP200_RADIUS  (0.545*0.5)
#define RMP200_COUNT1ROT (50000.0)

  double pi = 3.1415926535;

  //setting
  double wheelrL = status.wheelradiusL;
  double wheelrR = status.wheelradiusR;
  double radius = status.wheelwidth * 0.5;
  double count1rot = status.count1rot;

  if (status.odofirst && status.odogot) {
    status.int_left_prev  = status.int_left_wheel;
    status.int_right_prev = status.int_right_wheel;
    status.odofirst = false;
  }

  if (status.odofirst) {
    return false;
  }

  const static double meterpercountL = 2.0 * wheelrL * pi / count1rot;
  const static double meterpercountR = 2.0 * wheelrR * pi / count1rot;
  const static int iterate = 64;

  double countL = (double)(status.int_left_wheel  - status.int_left_prev);
  double countR = (double)(status.int_right_wheel - status.int_right_prev);
  

  countR /= (double)iterate;
  countL /= (double)iterate;

  double meterR = countR * meterpercountR;
  double meterL = countL * meterpercountL;
  

  double forward = (meterR + meterL)*0.5;
  double diffR = meterR - forward;

  double addtheta = diffR / radius;


  status.addx = 0.0;
  status.addy = 0.0;
  status.addtheta = 0.0;
  for (int i = 0; i < iterate; i++) {
    status.odoy += forward * cos(status.odotheta) * kxy;
    status.odox -= forward * sin(status.odotheta) * kxy;
    status.odotheta += addtheta * ktheta;

    status.addy += forward * kxy;
    status.addx += 0.0;
    status.addtheta += addtheta * ktheta;
  }

#define USE_ODOPRINT (0)
#if USE_ODOPRINT
  printf("odo %f %f %f enc   %d %d\n", 
	 (float)status.odox, (float)status.odoy, (float)status.odotheta, 
	 status.int_left_wheel, status.int_right_wheel);
#endif  

  status.int_left_prev  = status.int_left_wheel;
  status.int_right_prev = status.int_right_wheel;

  return true;
}






#if RMP_TEST01_NO_MEMMOVE
static void* movememory(void* dst, const void* src, size_t size)
{
	char* dstp = (char*) dst;
	char* srcp = (char*) src;
	for (int i = 0; i < size; i++) {
		*dstp = *srcp;
		dstp++;
		srcp++;		
	}
	return dst;
}
#else
static void* movememory(void* dst, const void* src, size_t size)
{
	return memmove(dst, src, size);
}

#endif

void rmp_getstatus(rmp_status& status, double kodoxy, double kodotheta)
{


   for (unsigned int loop = 0; loop < readloopcnt; loop++) {
      const int commandlen = 18;
      const int readmax = 2 * commandlen * 10;
      DWORD numread = 0;
      static unsigned char buf[readmax];
      static int storeidx = 0;
      static int lastidx = 0;

#if !RMP_TEST02_NO_FTDI
      usbs_Read(buf+storeidx, readmax-storeidx, &numread);
#endif     

#if 0
      printf("\n");
      for (int j = storeidx; j < storeidx+numread; j++) {
	if (buf[j] == 0xf0) {
	  printf("\n");
	}
	printf("%02x ", buf[j]);
      }
      printf("\n\n");
#endif



#define DBPRINT (0)

#if DBPRINT
      printf("readnum %d / %d\n", storeidx + (int)numread, readmax);
#endif

      if (storeidx + (int)numread - commandlen > 0) {
#if DBPRINT
	printf("inter %d\n", storeidx + (int)numread - commandlen);
#endif
	for (int i = 0; i <= storeidx + (int)numread - commandlen; i++) {

	  //use A-SIDE status
	  if (buf[i] == 0xF0 && buf[i+1] == 0x55 && buf[i+2] == 0xAA) {
	    //printf("found status\n");
	    rmp_getstatus_sub(&buf[i], status);
	    lastidx = i + commandlen;
	  }
	}
      }

      storeidx = storeidx + (int)numread - lastidx;
#if DBPRINT
      printf("lastidx %d  storeidx %d\n", lastidx, storeidx);
#endif
      movememory(&buf[0], &buf[lastidx], storeidx);
      lastidx = 0;

    }


   rmp_odometry(status, kodoxy, kodotheta);

   status.odotime_prev = status.odotime;
   status.odotime = status.tv403;
}


//===========================================
//SET SPEED
//===========================================
void rmp_setspeed(short speed, short turn, unsigned short command, unsigned short param)
{
  
    //===================================
    //write speed
    //===================================

    //write speed
    static unsigned char buf[18] = {
        //0   1     2     3     4     5     6     7     8
	0xf0, 0x55, 0x00, 0x00, 0x00, 0x00, 0x04, 0x13, 0x00,
	//9   10    11    12    13    14    15    16    17
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x70, 
    };

    //VEL
    buf[9]  = ((speed>>8) & 0xFF);
    buf[10] = ( speed     & 0xFF);
    //TURN
    buf[11] = ((turn>>8)  & 0xFF);
    buf[12] = ( turn      & 0xFF);

    buf[13] = ((command>>8) & 0xFF);
    buf[14] = ((command)    & 0xFF);
    buf[15] = ((param>>8)   & 0xFF);
    buf[16] = ((param)      & 0xFF);

#if 0
    buf[13] = 0x00;
    buf[14] = 0x00;//(0=NONE)(16=SET OPERATIONAL MODE)(50=RESET INTEGRATORS)
    buf[15] = 0x00;
    buf[16] = 0x02;//Balance Mode(in case of SET OPERATIONAL MODE)
#endif
#if !RMP_TEST02_NO_FTDI
    usbs_message_format(buf);
#endif

    DWORD written;

#if !RMP_TEST02_NO_FTDI
    usbs_Write(buf, 18, &written);
#endif
}


void rmp_setconfiguration(unsigned short command, unsigned short param)
{
    //write speed
    static unsigned char buf[18] = {
        //0   1     2     3     4     5     6     7     8
	0xf0, 0x55, 0x00, 0x00, 0x00, 0x00, 0x04, 0x13, 0x00,
	//9   10    11    12    13    14    15    16    17
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x70, 
    };

    //ConfigurationCommand
    buf[13] = ((command>>8) & 0xFF);
    buf[14] = ((command)    & 0xFF);
    buf[15] = ((param>>8)   & 0xFF);
    buf[16] = ((param)      & 0xFF);

    printf("rmp_if setconfiguration command %d param %d\n", command, param);

#if 0
    buf[13] = 0x00;
    buf[14] = 0x00;//(0=NONE)(16=SET OPERATIONAL MODE)(50=RESET INTEGRATORS)
    buf[15] = 0x00;
    buf[16] = 0x02;//Balance Mode(in case of SET OPERATIONAL MODE)
#endif

#if !RMP_TEST02_NO_FTDI
    usbs_message_format(buf);
#endif
    DWORD written;
#if !RMP_TEST02_NO_FTDI
    usbs_Write(buf, 18, &written);
#endif
}
